use std::fmt;
use std::error::Error as StdError;
use futures::IntoFuture;
use hyper::{Request, Method, Uri, Headers};
use hyper::header::{Header, ContentType};
use typemap::SendMap;
use tokio_core::reactor::{Handle, Remote};

use {HandleError, Outcome, Response, Responder};
use body::{Body, FromBody};
use router::regex::OwnedCaptures;


/// A context during handling.
///
/// It contains an HTTP request, HTTP response to return,
/// and a typemap in order to share variables between middlewares.
pub struct Context {
    pub method: Method,
    pub uri: Uri,
    /// Captured parameters from router
    pub params: OwnedCaptures,
    pub headers: Headers,
    /// Extension map to share values between handlers/middlewares
    pub ext: SendMap,
    /// Request body
    pub(crate) body: Option<Body>,
    /// Handle of event loop
    pub(crate) handle: Remote,
}

impl Context {
    pub fn from_components(req: Request, params: OwnedCaptures, handle: Remote) -> Self {
        let (method, uri, _http_version, headers, body) = req.deconstruct();
        Context {
            method,
            uri,
            params,
            headers,
            handle,
            body: Some(body.into()),
            ext: SendMap::custom(),
        }
    }

    /// Returns the path of request URL.
    pub fn path(&self) -> &str {
        self.uri.path()
    }

    pub fn header<T: Header>(&self) -> Option<&T> {
        self.headers.get::<T>()
    }

    pub fn respond<R: Responder>(self, responder: R) -> Outcome<Response> {
        Outcome::result(responder.respond(self))
    }

    pub fn take_body<T: FromBody>(&mut self) -> Outcome<T> {
        match self.headers.get::<ContentType>() {
            Some(&ContentType(ref m)) if T::check_mime(m) => (),
            _ => return Outcome::err(HandleError::bad_request()),
        }
        match self.body.take() {
            Some(body) => Outcome::boxed(body.into_value::<T>()),
            None => Outcome::err(HandleError::internal(EmptyBody)),
        }
    }

    /// Spawn a new future into the event loop.
    pub fn spawn<F, R>(self, f: F)
    where
        F: Send + 'static + FnOnce(&Handle) -> R,
        R: IntoFuture<Item = (), Error = ()>,
        R::Future: 'static,
    {
        self.handle.spawn(f)
    }
}



#[derive(Debug)]
struct EmptyBody;

impl fmt::Display for EmptyBody {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

impl StdError for EmptyBody {
    fn description(&self) -> &str {
        "empty body"
    }
}
