use std::io::{self, Read};
use std::fs::File as RawFile;
use std::path::{Path, PathBuf};
use hyper::Response as RawResponse;
use hyper::mime::{self, Mime};
use hyper::header::{Charset, ContentType, ContentDisposition, DispositionType, DispositionParam};
use {Context, HandleError, StatusCode};
use super::{Response, Responder};

/// A responder represents a file.
pub struct File {
    path: PathBuf,
    file: RawFile,
    media_type: Option<Mime>,
    is_attachment: bool,
    filename: Option<String>,
}

impl File {
    pub fn open<P: AsRef<Path>>(path: P) -> Result<Self, HandleError> {
        let path = path.as_ref().to_path_buf();
        let file = RawFile::open(&path).map_err(
            |err| if err.kind() ==
                io::ErrorKind::NotFound
            {
                HandleError::Status(StatusCode::NotFound)
            } else {
                HandleError::internal(err)
            },
        )?;
        Ok(File {
            path,
            file,
            media_type: None,
            is_attachment: false,
            filename: None,
        })
    }

    pub fn with_media_type(mut self, m: Mime) -> Self {
        self.media_type = Some(m);
        self
    }

    pub fn attachment(mut self, filename: Option<&str>) -> Self {
        self.is_attachment = true;
        self.filename = filename.map(ToOwned::to_owned);
        self
    }
}

impl Responder for File {
    fn respond(self, _: Context) -> Result<Response, HandleError> {
        let File {
            path,
            mut file,
            media_type,
            is_attachment,
            filename,
        } = self;

        let ext = path.extension().and_then(|s| s.to_str());
        let media_type = media_type.or_else(|| guess_mime(ext)).unwrap_or(
            mime::TEXT_PLAIN_UTF_8,
        );

        // TODO: use async read
        let mut content = Vec::new();
        file.read_to_end(&mut content).map_err(
            HandleError::internal,
        )?;

        let mut res = RawResponse::new()
            .with_header(ContentType(media_type))
            .with_body(content);

        if is_attachment {
            let filename = filename
                .as_ref()
                .map(|s| s.as_str())
                .or_else(|| path.file_name().and_then(|s| s.to_str()))
                .unwrap_or("content.txt");
            res.headers_mut().set(ContentDisposition {
                disposition: DispositionType::Attachment,
                parameters: vec![
                    DispositionParam::Filename(
                        Charset::Iso_8859_1,
                        None,
                        filename.as_bytes().into()
                    ),
                ],
            });
        }
        Ok(Response::from_raw(res))
    }
}

fn guess_mime(ext: Option<&str>) -> Option<Mime> {
    match ext {
        Some("json") => Some(mime::APPLICATION_JSON),
        Some("bmp") => Some(mime::IMAGE_BMP),
        Some("jpg") => Some(mime::IMAGE_JPEG),
        Some("png") => Some(mime::IMAGE_PNG),
        Some("html") => Some(mime::TEXT_HTML),
        Some("css") => Some(mime::TEXT_CSS),
        Some("js") => Some(mime::TEXT_JAVASCRIPT),
        Some("txt") => Some(mime::TEXT_PLAIN),
        Some("xml") => Some(mime::TEXT_XML),
        _ => None,
    }
}

#[test]
fn test_guess_mime() {
    assert_eq!(guess_mime(Some("html")), Some(mime::TEXT_HTML));
    assert_eq!(guess_mime(None), None);
}
