use std::any::Any;
use std::io;
use std::iter::Chain;
use std::sync::Arc;
use std::panic::AssertUnwindSafe;

use futures::{Future, BoxFuture};
use hyper::{self, StatusCode};
use hyper::server::{Service, NewService};
use tokio_core::reactor::{Handle, Remote};

use {Context, Outcome, Response, HandleError, Middleware};
use router::Router;


/// A base struct of `SusanooService`, which contains all components of services.
pub struct SusanooServiceFactoryBase {
    router: Router,
    middlewares: Vec<Arc<Middleware>>,
}

impl SusanooServiceFactoryBase {
    /// Create a new factory base from components
    pub fn new(router: Router, middlewares: Vec<Arc<Middleware>>) -> Self {
        SusanooServiceFactoryBase {
            router,
            middlewares,
        }
    }

    /// Create an instance of `SusanooServiceFactory`, associated with given event handle.
    pub fn create(&self, handle: &Handle) -> SusanooServiceFactory {
        SusanooServiceFactory {
            inner: SusanooService {
                handle: handle.remote().clone(),
                router: self.router.clone(),
                middlewares: self.middlewares.clone(),
            },
        }
    }
}


/// A factory of `SusanooService`
pub struct SusanooServiceFactory {
    inner: SusanooService,
}

impl NewService for SusanooServiceFactory {
    type Request = hyper::Request;
    type Response = hyper::Response;
    type Error = hyper::Error;
    type Instance = SusanooService;

    fn new_service(&self) -> io::Result<Self::Instance> {
        Ok(self.inner.clone())
    }
}


/// An asynchronous function executed by Tokio, associated with a HTTP request.
#[derive(Clone)]
pub struct SusanooService {
    handle: Remote,
    router: Router,
    middlewares: Vec<Arc<Middleware>>,
}

impl SusanooService {
    fn apply_handlers(&self, req: hyper::Request) -> Outcome<Response> {
        // Get a reference of matched route if exists
        let route = match self.router.find_route(req.method(), req.path()) {
            Some(route) => route,
            None => return Outcome::err(HandleError::Status(StatusCode::NotFound)),
        };

        // Build an instance of Context from req, route and self.handle
        let cap = route.pattern.owned_captures(req.path()).unwrap();
        let ctx = Context::from_components(req, cap, self.handle.clone());

        let middlewares = chain(&self.middlewares, &route.middlewares).map(Clone::clone);
        let handler = route.handler.clone();
        Outcome::boxed(
            middlewares
                .fold(Outcome::ok(ctx).boxed(), |ctx, m| {
                    ctx.and_then(move |ctx| m.call(ctx)).boxed()
                })
                .and_then(move |ctx| handler.call(ctx)),
        )
    }

    // TODO: use custom error handler,
    fn handle_error(err: HandleError) -> hyper::Response {
        match err {
            HandleError::Raw(res) => res,
            HandleError::Status(status) => hyper::Response::new().with_status(status),
            HandleError::Internal(err) => {
                hyper::Response::new()
                    .with_status(StatusCode::InternalServerError)
                    .with_body(format!("Internal server error: {:?}", err))
            }
        }
    }

    fn into_raw_response(
        outcome: Result<Result<Response, HandleError>, Box<Any + Send>>,
    ) -> hyper::Response {
        match outcome {
            Ok(Ok(res)) => res.into_raw(),
            Ok(Err(err)) => Self::handle_error(err),
            Err(err) => {
                hyper::Response::new()
                    .with_status(StatusCode::InternalServerError)
                    .with_body(format!("Internal server error: panic: {:?}", err))
            }
        }
    }
}

impl Service for SusanooService {
    type Request = hyper::Request;
    type Response = hyper::Response;
    type Error = hyper::Error;
    type Future = BoxFuture<Self::Response, Self::Error>;

    fn call(&self, req: Self::Request) -> Self::Future {
        let outcome = AssertUnwindSafe(self.apply_handlers(req)).catch_unwind();
        outcome
            .then(|outcome| Ok(Self::into_raw_response(outcome)))
            .boxed()
    }
}

fn chain<I, J>(i: I, j: J) -> Chain<I::IntoIter, J::IntoIter>
where
    I: IntoIterator,
    J: IntoIterator<Item = I::Item>,
{
    i.into_iter().chain(j.into_iter())
}
