use std::sync::Arc;
use std::net::SocketAddr;

use hyper::server::Http;
use tokio_proto::TcpServer;

use Middleware;
use router::{Router, Route};
use super::server::SusanooServiceFactoryBase;

/// Root instance of Web application.
#[derive(Default)]
pub struct Susanoo {
    listen_address: Option<SocketAddr>,
    num_threads: Option<usize>,
    router: Router,
    middlewares: Vec<Arc<Middleware>>,
}

impl Susanoo {
    /// Add a route to the application.
    pub fn with_route(mut self, route: Route) -> Self {
        self.router.add_route(route);
        self
    }

    /// Add a middleware to the application.
    ///
    /// If you want to apply middelewares to certain route, use `Route::with_middleware`.
    pub fn with_middleware<M: Middleware>(mut self, middleware: M) -> Self {
        self.middlewares.push(Arc::new(middleware));
        self
    }

    /// Set listen address of the application.
    ///
    /// The default address is `"0.0.0.0:4000"`.
    pub fn bind<S: Into<SocketAddr>>(mut self, addr: S) -> Self {
        self.listen_address = Some(addr.into());
        self
    }

    /// Set the number of threads running simulateneous event loops (Unix only).
    ///
    /// The default value is 1.
    #[cfg(not(windows))]
    pub fn threads(mut self, num: usize) -> Self {
        self.num_threads = Some(num);
        self
    }

    /// Start the server with given configuration.
    ///
    /// This method will block the current thread until the server is shut down.
    pub fn run(self) {
        let listen_address = self.listen_address.unwrap_or_else(|| {
            "0.0.0.0:4000".parse().unwrap()
        });
        let num_threads = self.num_threads.unwrap_or(1);

        let mut server = TcpServer::new(Http::new(), listen_address);
        server.threads(num_threads);

        let factory = SusanooServiceFactoryBase::new(self.router, self.middlewares);
        server.with_handle(move |handle| factory.create(handle));
    }
}
